# Generated by Django 2.2.14 on 2020-07-26 18:18

from django.db import migrations, models
import django.db.models.deletion
import django.db.models.manager


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='AbstractVegetable',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(choices=[('testapp.fruit', 'fruit'), ('testapp.vegetable', 'vegetable')], db_index=True, max_length=255)),
                ('name', models.CharField(max_length=255)),
                ('color', models.CharField(max_length=255)),
                ('yumness', models.FloatField()),
            ],
            options={
                'abstract': False,
            },
            managers=[
                ('mymanager', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='Animal',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(choices=[('testapp.angrybigcat', 'angry big cat'), ('testapp.bigcat', 'big cat'), ('testapp.canine', 'canine'), ('testapp.feline', 'feline'), ('testapp.parrot', 'parrot')], db_index=True, max_length=255)),
                ('name', models.CharField(max_length=255)),
                ('mice_eaten', models.IntegerField(default=0)),
                ('known_words', models.IntegerField(null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UniqueIdentifier',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField(blank=True, null=True)),
                ('created', models.DateTimeField(auto_now_add=True, db_index=True)),
                ('name', models.CharField(max_length=255)),
                ('content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='Parent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.CharField(choices=[('testapp.child1', 'child1'), ('testapp.child2', 'child2')], db_index=True, max_length=255)),
                ('a', models.CharField(max_length=1)),
                ('b', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='testapp.Parent')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Canine',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.animal',),
        ),
        migrations.CreateModel(
            name='Child1',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.parent',),
        ),
        migrations.CreateModel(
            name='Child2',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.parent',),
        ),
        migrations.CreateModel(
            name='Feline',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.animal',),
        ),
        migrations.CreateModel(
            name='Fruit',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.abstractvegetable',),
            managers=[
                ('mymanager', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='Parrot',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.animal',),
        ),
        migrations.CreateModel(
            name='Vegetable',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.abstractvegetable',),
            managers=[
                ('mymanager', django.db.models.manager.Manager()),
            ],
        ),
        migrations.AddField(
            model_name='animal',
            name='canines_eaten',
            field=models.ManyToManyField(to='testapp.Canine'),
        ),
        migrations.CreateModel(
            name='BigCat',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.feline',),
        ),
        migrations.CreateModel(
            name='AngryBigCat',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('testapp.bigcat',),
        ),
    ]
